/*
**      c p p _ e p o c . c
**
**  This module handles the interface to the operating system.
**
** C compiler
** ==========
**
** Copyright 1995,  K.D.Walker
***             and D.J.Walker
**
** History:
**
** Feb 1995     The operating and environment dependencies isolated
**              in this module.  This has been done primarily as part
**              of the preprocessor to run on the EPOC operating system
**              used by the Psion 3a pocket computer.
**/

#define EPOC_BUFSIZE 512

#ifdef TESTING
#define DBG(params)  fprintf params
#else
#define DBG(params)
#endif

#if 0
#include "cpp.h"
#endif

/******************************* EPOC ***********************************/
/*
 *  The EPOC operating system is used on the Psion range of pocket
 *  computers.   These systems are based on the Intel X86 range of
 *  processors.  All programs have to confrom to the small memory
 *  model (64k code + 64k data), although there are facilities for
 *  using inter-process communication and Dynamic Link Libraries to
 *  circumvent some of these restrictions.
 *
 *  The Psion SDK does provide a library (CLIB) that provides full
 *  support for all ANSI features, including those that would be
 *  required to use the DEFAULT section above.  However, if this is
 *  done, the compiler is too large to fit within the constraints of
 *  the small memory model.
 *
 *  An alternative library (PLIB) is available that is instead based
 *  around calling functions that are held in the Psion ROM.  This
 *  provides significantly smaller executables than are generated
 *  when CLIB is used.  However the functionality match to the ANSI
 *  routines is not exact, so that the emulation routines defined in
 *  this section are used instead.  Doing this, it is possible (just)
 *  to get the compiler within the small memory model.
 */

#ifdef EPOC

#include <stdarg.h>
#include <errno.h>
#include <stddef.h>

#define CPOC_STDIO
#include <cpoclib.h>

CPOC_FILE * stdin = NULL;
CPOC_FILE * stdout = NULL;
CPOC_FILE * stderr = NULL;

#define FILE CPOC_FILE

static void do_initfp (FILE ** fp, VOID *ptr)
{
    (*fp)->IOptr = ptr;
    (*fp)->IOend = ptr;
    (*fp)->IObuf = ptr;
}
/*
 *  We have a very simple fopen() as we only have to cater for
 *  simple read/write modes.  The one optimisation we do make
 *  is to allocate an I/O buffer for input files to avoid
 *  excessive calls to the EPOC file handler as these are expensive
 *  in processing terms.
 */
static FILE * do_fopen (const char *name, const char * mode, FILE *fp)
{
    char * ptr;
    unsigned char * bufptr;

    /*
     *  Convert any forward slashes in the name
     *  into backslashes for EPOC
     */
    for (ptr = (char *)name ; *ptr ; ptr++) {
        if (*ptr == '/') {
            *ptr = '\\';
        }
    }
    bufptr = NULL;
    /*
     *  Open according to mode
     */
    switch (*mode) {
    case 'a':      /* Append if exists with write access */
        if (p_open(&fp->handle,(char *)name,
                                P_FSTREAM_TEXT | P_FUPDATE | P_FAPPEND) == 0) {
            break;
        }
        /* FALL THRU */
    case 'w':     /* Create new file with write access */
        if (p_open(&fp->handle,(char *)name,
                                P_FSTREAM_TEXT | P_FUPDATE | P_FREPLACE) == 0) {
            break;
        }
        return (NULL);
    case 'r':     /* Open existing file with READ access */
        bufptr = p_alloc(EPOC_BUFSIZE + 1);
        if (bufptr == NULL) {
            p_panic(150);
        }
        if (p_open(&fp->handle, (char *)name,
                                P_FSTREAM_TEXT | P_FOPEN) < 0) {
            p_free(bufptr);
            return (NULL);
        }
        break;
    default:
        p_exit(5);
    }
    do_initfp (&fp,bufptr);
    return (fp);
}

FILE * fopen (const char *name, const char *mode)
{
    FILE * fp;

    fp = p_alloc (sizeof (FILE));
    if (fp == NULL) {
        p_panic(151);
    }
    if (do_fopen(name,mode,fp) == NULL) {
        p_free(fp);
#ifdef TESTING
        if (stderr != NULL)
            DBG ((stderr,"fopen(\"%s\",\"%s\") failed\n",name,mode));
#endif
       return (NULL);
    }

#ifdef TESTING
if (stderr != NULL)
    DBG ((stderr,"fopen(\"%s\",\"%s\",fp=$%x), IObuf=$%x\n",name,mode,fp,fp->IObuf));
#endif

    return (fp);
}

/*
 *  Support routine for fclose() and freopen()
 */
static void do_fclose( FILE * fp)
{
    if (fp != NULL) {
       if (fp->handle != NULL) {
           p_close (fp->handle);
           fp->handle = NULL;
       }
       if (fp->IObuf != NULL) {
           p_free (fp->IObuf);
           fp->IObuf = NULL;
       }
    }
    return;
}

FILE * freopen (const char *name, const char *mode, FILE *fp)
{
    do_fclose(fp);
    fp = do_fopen(name,mode,fp);

#ifdef TESTING
    if (fp != NULL)
        DBG((stderr, "freopen(\"%s\",\"%s\",fp=$%x), IObuf=$%x\n",name,mode,fp,fp->IObuf));
#endif

    return (fp);
}

int  fclose (FILE * fp)
{
    if (fp != NULL) {
        do_fclose(fp);
        p_free (fp);
    }
    return 0;
}

/*
 *  The routine build into the EPOC ROM is very similar to the
 *  ANSI C fread() routine except that it returns a -ve value on
 *  error rather than 0, and we need to correct for this
 *
 *  However for effeciency reasons we also want to read in blocks
 *  of 512 characters, so we implement our own buffering here!
 */
static
size_t
fread (char* buffer, size_t size, size_t num, FILE * fp)
{
    size_t count;
    size_t copied;
    int   reply;

    for (count = size * num, copied = 0 ; copied < count ;
                         buffer++, copied++, fp->IOptr++) {
        if (fp->IOptr >= fp->IOend) {
            if (fp->handle == NULL)
               p_panic(152);
            if (fp->IObuf == NULL)
               p_panic(153);
            reply = p_read (fp->handle, fp->IObuf, EPOC_BUFSIZE);
            DBG((stderr,"p_read: fp=$%x, reply=%d\n", fp, reply));
            if (reply <= 0) {
                break;
            }
            fp->IOend = &fp->IObuf[reply];
            fp->IOptr = &fp->IObuf[0];
        }
        buffer[0] = fp->IOptr[0];
    }
    return (copied);
}

char * fgets (char * buf, int count, FILE * fp)
{
    int   n;
    char * s;

    for (n=0, count--, s = buf ; n <= count ; n++, s++ ) {
        if (fp->IOptr < fp->IOend) {
            *s = *fp->IOptr++;
        } else {
            if (fread(s,1,1,fp) <= 0) {
                break;
            }
        }
        if (*s == '\n') {
            n++;
            s++;
            break;
        }
        if (*s == '\r') {
            /* ignore CR characters */
            s--;
            n--;
            continue;
        }
    }
EXITFOR:
    *s = '\0';
    if (n == 0)  {
        DBG((stderr,"fgets(fp=$%x):  EOF found\n", fp));
        return NULL;
    }
    DBG((stderr,"fgets(fp=$%x, count=%d(n=%d), buf=%s",fp, count, n, buf));
    return buf;
}


/*
 *  EPOC based systems have a ROM based routine that is equivalent
 *  to a C style sprintf routine.   We take advantage of this to
 *  avoid the need to pull in a STDIO package.
 *
 *  The following important assumptions apply:
 *
 *  a)  The return value from 'vfprintf' is only relevant so far
 *      as it is checked as being non-zero
 *
 *  b)  The output must not exceed the size of the buffer allocated below
 *      N.B.  Should this be a constant set in chdr.h?
 *
 *  We also (potentially) make the assumption that for stderr
 *  we will do a write to the parent job instead.
 */

int
vfprintf (FILE *fp, const char * fmt, va_list args)
{
    int     len,reply;

    if (fp == NULL && fp != stderr)
        p_panic(154);
    len = p_atob(&_vfprintf_buf[0], (char *)fmt, args);

    if (len > _vfprintf_len-1)
        p_panic(158);
    if (fp == stderr) {
       if (WriteToParentSetup() != NULL) {
           _vfprintf_buf[len] = '\0';
           WriteToParent(&_vfprintf_buf[0], len+1);
           return (len);
       }
    }
    reply = p_write (fp->handle, &_vfprintf_buf[0], len);
    return (reply);
}

/*
 *  The 'errno' variable is not defined on EPOC based systems when
 *  using the PLIB libraary.
 */
void perror ( const char *s)
{
    if (s && *s)
        fprintf (stderr,"%s: ",s);
    fprintf(stderr,"%s\n","");
}

/*
 *  On EPOC based systems there is no stderr file opened by default
 *  when using PLIB.  We will therefore try and open such a channel
 *  on the current default drive.
 *
 *  We will try and append to the file if it already exists.  This
 *  allows for batch compiles.  It is up to the CPOC front-end to
 *  delete the error file as required.
 */
void
openerror (void)
{
    stdin = p_alloc (sizeof (FILE));
    stdout = p_alloc (sizeof (FILE));
    do_initfp (&stdin, NULL);
    do_initfp (&stdout, NULL);
    _openerror ();
}

#endif /* EPOC */

